{*******************************************************************************
*                                                                              *
*  von [DC]ObeY - www.dunecoop.de - obey@dunecoop.de ;-)                       *
*                                                                              *
*******************************************************************************}

unit WRL;

interface

uses
	Windows,  // fr TBitmap - muss vorher der Unit Graphics stehen
	Messages, 
	SysUtils, // fr TException
	Classes,  // Fr TFileStream
	Graphics, // fr TColor
	Controls, 
	Forms, 
	extctrls, // fr TImage
	Dialogs;

type
	EMyError = class(Exception);

	TWRL = class ( TComponent )
  private
		{ Private-Deklarationen }
		ipvBreite : Integer; // Anzahl Spielfelder horizontal
		ipvHoehe : Integer; // Anzahl Spielfelder vertikal
		apvMiniMap : array of byte; // Minimap mit Index auf Palette
		apvTiles : array of integer; // Spielfeld, verweis auf Terrains
		ipvTerrainC : Integer; // Anzahl Terrains AT = 2 bytes
		apvTerrainUse : Array of Integer; // fr wieviele Tiles wird jedes Terrain genutzt?
		apvTerrain : Array of byte; // Terrains AT * 64 * 64
		apvPalette : Array of TColor; // 256 Farben * 3 bytes
		apvTerrainKind : Array of byte; // Art Terrain / 0 Land / 1 Wasser / 2 Kste / 3 unpassierbar
		apvTerrainImages : Array of TBitmap; // Alle Terrainbilder als Bitmap
		bpoMiniMap : TBitmap; // MiniMap als Bitmap
		zpvFile : String; // Dateiname inkl. Pfad
		ipvBytes : Int64; // Dateigre in Bytes (mitgezhlt)
		ipvFileSize : Int64; // Dateigre ausgelesen

		function GetMiniMap : TBitmap; // MiniMap Bild
		function GetPaletteColor ( Index : Integer ) : TColor; // TColor der Palette ber Index
		function GetTerrainUse ( Index : Integer ) : Integer; // Wie oft wurde ein Terrain verwendet?
		function GetPaletteCount : Integer; // Anzahl Elemente im Palettenarray
		function GetMiniMapColor ( xIndex : Integer; yIndex : Integer ) : TColor;
		function GetPalette : HPALETTE;
		function GetTileCount : Integer; // gibt die Anzahl der Tiles zurck (Breite * Hhe)
		
	protected
		{ Protected-Deklarationen }
					
	public
		{ Public-Deklarationen }
		constructor Create( AOwner : TComponent); Override; // (FileName : String); 
		destructor Destroy(); override;
		
		procedure LoadWRLFile ( FileName : String);

		Property Width : Integer read ipvBreite; // Breite der Karte
		Property Heigth : Integer read ipvHoehe; // Hhe der Karte
		function GetTile ( xIndex : Integer; yIndex : Integer ) : TBitmap; // Kartenkachel unter Angabe von X und Y
		function GetTileIndex ( xIndex : Integer; yIndex : Integer ) : Integer; // gibt die Referenz auf das Terrain Array zurck
		Property TileCount : Integer read GetTileCount;
		// MiniMap
		property MiniMap: TBitMap read GetMiniMap; // MiniMap als BitMap
		Property MiniMapColor [ xIndex : Integer; yIndex : Integer ] : TColor read GetMiniMapColor; // Pixelwert der Minimap
		// Farbpalette
		property PaletteColor [ Index : Integer ] : TColor read GetPaletteColor; // Farbwert der Palette (Index 0 - 255 )
		property PaletteColorCount : Integer read GetPaletteCount; // Anzahl Farbwerte in der Palette / immer 256 ?!
		// Terrain
		property TerrainCount : Integer read ipvTerrainC; // Anzahl der verschiedenen Terrains
		function GetTerrain ( Index : Integer ) : TBitmap; // Terrainbild von Index
		property TerrainUse [ Index : Integer ] : Integer read GetTerrainUse;
		
	published
    { Published-Deklarationen }
  end;

procedure Register;

implementation

const
	bpaHeader: array[0..4] of byte = (87, 82, 76, 1, 0);

procedure Register;
begin
	RegisterComponents('MAXR', [TWRL]);
end;

{*******************************************************************************
*                                                                              *
*                                                                              *
*                                                                              *
*******************************************************************************}
function GetFileSize(const szFile: String): Int64;
var
  fFile: THandle;
  wfd: TWIN32FINDDATA;
begin
  result := 0;
  if not FileExists(szFile) then exit;
  fFile := FindFirstfile(pchar(szFile),wfd);
	if fFile = INVALID_HANDLE_VALUE then exit;
  result := (wfd.nFileSizeHigh*(MAXDWORD))+wfd.nFileSizeLow;
  windows.FindClose(fFile);
end; 
	
{*******************************************************************************
*                                                                              *
*                                                                              *
*                                                                              *
*******************************************************************************}
function ArraysEqual(const Array1, Array2: array of byte): Boolean;
begin
	Result := Length(Array1) = Length(Array2);
	if Result and (Length(Array1) > 0) then
		Result := CompareMem(@Array1, @Array2, SizeOf(Array1[0]) * Length(Array1));
end;	

{*******************************************************************************
*                                                                              *
*  convert hexa-decimal values to RGB                                          *
*                                                                              *
*******************************************************************************}
function RGB2TColor(const R, G, B: Byte): Integer;
begin
	Result := R + G shl 8 + B shl 16;
end;

{*******************************************************************************
*                                                                              *
*  convert hexa-decimal values to RGB                                          *
*                                                                              *
*******************************************************************************}
procedure TColor2RGB(const Color: TColor; var R, G, B: Byte);
begin
	
	R := Color and $FF;
	G := (Color shr 8) and $FF;
	B := (Color shr 16) and $FF;
end; 

{*******************************************************************************
*                                                                              *
*                                                                              *
*                                                                              *
*******************************************************************************}
function TWRL.GetPalette : HPALETTE;
var                       
	pal: PLogPalette;
	hpal: HPALETTE;
	i: Integer;
	r,g,b : byte;
	
begin
	Result := MiniMap.Palette; // unterdrckt Warnung des Compilers
	pal := nil;
	try
		GetMem(pal, sizeof(TLogPalette) + sizeof(TPaletteEntry) * 255);
		pal.palVersion := $300;
		pal.palNumEntries := 256;
		for i := 0 to 255 do
		begin
			TColor2RGB(apvPalette[i],r,g,b);
			pal.palPalEntry[i].peRed := r;
			pal.palPalEntry[i].peGreen := g;
			pal.palPalEntry[i].peBlue := b;
		end;
		hpal := CreatePalette(pal^);
		if hpal <> 0 then 
			Result := hpal;
	finally
		FreeMem(pal);
	end;
end;

{*******************************************************************************
*                                                                              *
*                                                                              *
*                                                                              *
*******************************************************************************}
function TWRL.GetTileCount : Integer;
Begin
	Result := Width * Heigth;
End;

{*******************************************************************************
*                                                                              *
*  Wie oft wurde ein Terrain verwendet?                                        *
*                                                                              *
*******************************************************************************}
function TWRL.GetTerrainUse ( Index : Integer ) : Integer; 
Begin
	Result := apvTerrainUse[Index];
End;

{*******************************************************************************
*                                                                              *
*                                                                              *
*                                                                              *
*******************************************************************************}
function TWRL.GetTerrain ( Index : Integer ) : TBitmap; 
Begin
	Result := apvTerrainImages[Index];
End;

{*******************************************************************************
*                                                                              *
*                                                                              *
*                                                                              *
*******************************************************************************}
function TWRL.GetMiniMap : TBitmap;
Begin
	Result := bpoMiniMap;
End;

{*******************************************************************************
*                                                                              *
*                                                                              *
*                                                                              *
*******************************************************************************}
function TWRL.GetPaletteColor ( Index : Integer ) : TColor;
Begin
	Result := apvPalette[Index];
End;

{*******************************************************************************
*                                                                              *
*                                                                              *
*                                                                              *
*******************************************************************************}
function TWRL.GetPaletteCount : Integer;
Begin
	Result := Length(apvPalette);
end;


{*******************************************************************************
*                                                                              *
*                                                                              *
*                                                                              *
*******************************************************************************}
function TWRL.GetMiniMapColor ( xIndex : Integer; yIndex : Integer ) : TColor;
Begin
	Result := apvPalette[apvMiniMap[xIndex+yIndex*ipvHoehe]];
End;

{*******************************************************************************
*                                                                              *
*                                                                              *
*                                                                              *
*******************************************************************************}
function TWRL.GetTile ( xIndex : Integer; yIndex : Integer ) : TBitmap;
Begin
	Result := apvTerrainImages[apvTiles[xIndex+yIndex*ipvBreite]];
End;

{*******************************************************************************
*                                                                              *
*                                                                              *
*                                                                              *
*******************************************************************************}
function TWRL.GetTileIndex ( xIndex : Integer; yIndex : Integer ) : Integer;
Var
	i : Integer;

Begin
	i := apvTiles[xIndex + yIndex * ipvBreite];
	If i > ipvTerrainC Then
		ShowMessage ( 'GetTileIndex: i ist zu gro ' + IntToStr(i) );
	Result := i;
End;


{*******************************************************************************
*                                                                              *
*  Constructor                                                                 *
*                                                                              *
*******************************************************************************}
constructor TWRL.Create( AOwner  : TComponent );
 //(FileName : String); 
Begin
	inherited Create(AOwner); // hierdurch wird der ursprngliche Konstruktor aufgerufen
	bpoMiniMap := TBitmap.Create();
	bpoMiniMap.PixelFormat := pf8bit;
//	if FileExists(FileName) Then
//		LoadWRLFile(FileName);
End;


{*******************************************************************************
*                                                                              *
*  Destructor                                                                  *
*                                                                              *
*******************************************************************************}
destructor TWRL.Destroy();
Begin
	bpoMiniMap.Destroy();
	inherited;
end;


{*******************************************************************************
*                                                                              *
*  Ld eine WRL Datei                                                          *
*                                                                              *
*******************************************************************************}
procedure TWRL.LoadWRLFile ( FileName : String);
Var
	floFStream : TFileStream;
	alaArbe : Array of byte;
	i,j,k : Integer;
	ilvArbe : Integer;
	ilvArb2 : Integer;
	ilvArb3 : Integer;

Begin

	zpvFile := FileName;
	ipvBytes := 0;
	floFStream := TFileStream.Create(FileName, fmOpenRead); // Dateistream ffnen
	Try
		// Header                                                       
		SetLength(alaArbe,5);
		floFStream.ReadBuffer(alaArbe[0],5);                             { 5 bytes }
		Inc(ipvBytes,5);
		If Not(CompareMem(alaArbe,@bpaHeader,sizeof(alaArbe[0])*Length(alaArbe))) Then
			ShowMessage('Keine WRL Datei');

			
		// Kartengre
		// Breite
		SetLength(alaArbe,2);                                           
		floFStream.ReadBuffer(alaArbe[0],2);                             { 2 bytes }
		Inc(ipvBytes,2);
		ipvBreite := alaArbe[0] + alaArbe[1] * 256;                     
		// Hhe
		floFStream.ReadBuffer(alaArbe[0],2);                             { 2 bytes }
		Inc(ipvBytes,2);
		ipvHoehe := alaArbe[0] + alaArbe[1] *256;
		
		// Minimap - Verweis auf Palette
		SetLength(apvMiniMap,ipvBreite * ipvHoehe);
		floFStream.ReadBuffer(apvMiniMap[0],High(apvMiniMap)+1);       { x*y bytes }
		Inc(ipvBytes,High(apvMiniMap)+1);
		
		// Kacheldaten breite * hhe * 2 bytes
		SetLength(apvTiles,ipvBreite * ipvHoehe); // Array apvTiles Integer
		SetLength(alaArbe,2); // Hilfsarrays
		for i := 1 to ipvBreite * ipvHoehe Do
		Begin
			floFStream.ReadBuffer(alaArbe[0],2);                       { x*y*2 bytes }
			Inc(ipvBytes,2);
			apvTiles[i-1] := alaArbe[0] + alaArbe[1] *256;
		end;
		
		// Anzahl Terrains
		floFStream.ReadBuffer(alaArbe[0],2);                             { 2 bytes }
		Inc(ipvBytes,2);
		ipvTerrainC := alaArbe[0] + alaArbe[1] *256;
			
		// Terrain Daten AT * 64 * 64
		SetLength(apvTerrain,ipvTerrainC * 64 * 64); // Array apvTerrain byte
		floFStream.ReadBuffer(apvTerrain[0],High(apvTerrain)+1);   { c*64*64 bytes }
		Inc(ipvBytes,High(apvTerrain)+1);
		
		// Palette 256 * 3;
		SetLength(alaArbe,3); // Hilfsarrays
		SetLength(apvPalette,256); // Palette fr Farbwerte
		For i := 1 to 256 Do
		Begin
			// RGB = TColor
			floFStream.ReadBuffer(alaArbe[0],3);                       { 3*256 bytes }
			Inc(ipvBytes,3);
			apvPalette[i-1] := StrToInt('$00' + IntToHex(alaArbe[2],2) + 
												 IntToHex(alaArbe[1],2) + IntToHex(alaArbe[0],2)); 
		End;
		
		// Art Terrain
		SetLength(apvTerrainKind,ipvTerrainC); // Terrainart fr alle Terrains
		floFStream.ReadBuffer(apvTerrainKind[0],ipvTerrainC);              { TC bytes }
		Inc(ipvBytes,ipvTerrainC);


		{                         FileStream - Ende                                }

		// fr wieviele Tiles wird jedes Terrain genutzt?
		SetLength(apvTerrainUse,ipvTerrainC); 
		For i := 0 To High(apvTerrainUse) Do // mit 0 initialisieren
			apvTerrainUse[i] := 0;    

		// Zhle wie oft welches Tile verwendet wird
		// Alle Tiles durchgehen
		k := 0;
		For i := 0 To High(apvTiles) Do 
		Begin
			k := k + 1;
			j := apvTiles[i];
			If ( j < 0 ) or (j > ( ipvTerrainC-1 )) Then
				ShowMessage ( 'j zu gro ' + IntToStr(j) );
			Inc(apvTerrainUse[j]);
		End;     
//		ShowMessage ( IntToStr(k) );
		
		
		// Dateigre
		ipvFileSize := GetFileSize(zpvFile);
//		If ipvBytes <> ipvFileSize Then
//			ShowMessage ( 'Format falsch! Abweichung in Dateigre' );
		
		// Bitmap der MiniMap mit Palette initialisieren und danach Pixel einfrben
		MiniMap.Palette := GetPalette();
		bpoMiniMap.Width := ipvBreite;
		bpoMiniMap.Height := ipvHoehe;
		For j := 0 To ipvBreite - 1 Do
			For k := 0 To ipvHoehe - 1 Do
				bpoMiniMap.Canvas.Pixels[j,k] := apvPalette[apvMiniMap[j+k*ipvHoehe]];   

		// Alle mglichen Terrains als TBitmap
		SetLength(apvTerrainImages,ipvTerrainC);
		ilvArb2 := 0;
		For i := 1 To ipvTerrainC Do
		Begin
			apvTerrainImages[i-1] := Graphics.TBitmap.Create();
			apvTerrainImages[i-1].Width := 64;
			apvTerrainImages[i-1].Height := 64;
			apvTerrainImages[i-1].PixelFormat := pf8bit;
			apvTerrainImages[i-1].Palette := GetPalette();
			
			// Farbwerte zuweisen
			For j := 0 To 63 Do
				For k := 0 To 63 Do   
				Begin
					ilvArbe := k + j * 64 + (i-1) * 64 * 64;
					if ilvArb2 + 1 = ilvArbe Then
						ShowMessage ( 'fehler' )
					Else
						Inc(ilvArb2);
					ilvArb3 := apvTerrain[ilvArbe];
					if ilvArb3 > 255 Then
						ShowMessage ( 'ilvArb3 > 255 ' + IntToStr(ilvArb3) );
					apvTerrainImages[i-1].Canvas.Pixels[k,j] := apvPalette[ilvArb3];   
				End;
		End;                                 
		// image1.picture.bitmap.scanline

	finally
		alaArbe := Nil;
		if Not(floFStream = Nil) Then
			floFStream.Free;	
	End;


End;

end.
 